@@ -1,7 +1,7 @@
 /*
  * print.c	Routines to print stuff.
  *
- * Version:	$Id$
+ * Version:	$Id: e6bb49e821eebbe04eaf8cbb4c3c2c2c738e84da $
  *
  *   This library is free software; you can redistribute it and/or
  *   modify it under the terms of the GNU Lesser General Public
@@ -21,7 +21,7 @@
  */
 
 #include	<freeradius-devel/ident.h>
-RCSID("$Id$")
+RCSID("$Id: e6bb49e821eebbe04eaf8cbb4c3c2c2c738e84da $")
 
 #include	<freeradius-devel/libradius.h>
 
@@ -35,7 +35,7 @@ RCSID("$Id$")
  *	Note that we don't care about the length of the input string,
  *	because '\0' is an invalid UTF-8 character.
  */
-static int utf8_char(uint8_t *str)
+int fr_utf8_char(const uint8_t *str)
 {
 	if (*str < 0x20) return 0;
 
@@ -127,9 +127,10 @@ static int utf8_char(uint8_t *str)
  *	has to be larger than the input string by at least 5 bytes.
  *	If not, the output is silently truncated...
  */
-void fr_print_string(const char *in, size_t inlen, char *out, size_t outlen)
+size_t fr_print_string(const char *in, size_t inlen, char *out, size_t outlen)
 {
-	uint8_t		*str = (uint8_t *)in;
+	const char	*start = out;
+	const uint8_t	*str = (const uint8_t *) in;
 	int		sp = 0;
 	int		utf8 = 0;
 
@@ -144,7 +145,7 @@ void fr_print_string(const char *in, size_t inlen, char *out, size_t outlen)
 		 *	Some clients send strings with an off-by-one
 		 *	length (confused with strings in C).
 		 */
-		if ((inlen == 0) && (*str == 0)) break;
+		if ((inlen == 1) && (*str == 0)) break;
 
 		switch (*str) {
 			case '\\':
@@ -176,7 +177,7 @@ void fr_print_string(const char *in, size_t inlen, char *out, size_t outlen)
 			continue;
 		}
 
-		utf8 = utf8_char((uint8_t *)str);
+		utf8 = fr_utf8_char(str);
 		if (!utf8) {
 			snprintf(out, outlen, "\\%03o", *str);
 			out  += 4;
@@ -193,6 +194,8 @@ void fr_print_string(const char *in, size_t inlen, char *out, size_t outlen)
 		} while (--utf8 > 0);
 	}
 	*out = 0;
+
+	return out - start;
 }
 
 
@@ -283,7 +286,7 @@ int vp_prints_value(char * out, size_t outlen, VALUE_PAIR *vp, int delimitst)
 		case PW_TYPE_ABINARY:
 #ifdef ASCEND_BINARY
 			a = buf;
-			print_abinary(vp, buf, sizeof(buf));
+			print_abinary(vp, buf, sizeof(buf), delimitst);
 			break;
 #else
 		  /* FALL THROUGH */
@@ -475,7 +478,9 @@ void vp_print(FILE *fp, VALUE_PAIR *vp)
 	char	buf[1024];
 
 	vp_prints(buf, sizeof(buf), vp);
+	fputc('\t', fp);
 	fputs(buf, fp);
+	fputc('\n', fp);
 }
 
 
@@ -486,9 +491,7 @@ void vp_print(FILE *fp, VALUE_PAIR *vp)
 void vp_printlist(FILE *fp, VALUE_PAIR *vp)
 {
 	for (; vp; vp = vp->next) {
-		fprintf(fp, "\t");
 		vp_print(fp, vp);
-		fprintf(fp, "\n");
 	}
 }
 
