@@ -1,7 +1,7 @@
 /*
  * radiusd.c	Main loop of the radius server.
  *
- * Version:	$Id: 8682e63d3eafbd7e07acbee3e06f112357e4710b $
+ * Version:	$Id: f29c92809ab0cd50eda54315ae7220d8312b6101 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: 8682e63d3eafbd7e07acbee3e06f112357e4710b $")
+RCSID("$Id: f29c92809ab0cd50eda54315ae7220d8312b6101 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
@@ -53,6 +53,10 @@ RCSID("$Id: 8682e63d3eafbd7e07acbee3e06f112357e4710b $")
 #	define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
 #endif
 
+#ifdef HAVE_OPENSSL_CRYPTO_H
+#include <openssl/ssl.h>
+#endif
+
 /*
  *  Global variables.
  */
@@ -263,14 +267,6 @@ int main(int argc, char *argv[])
 		}
 	}
 
-	/*
-	 *	Mismatch between build time OpenSSL and linked SSL,
-	 *	better to die here than segfault later.
-	 */
-	if (ssl_check_version() < 0) {
-		exit(1);
-	}
-
 	if (flag && (flag != 0x03)) {
 		fprintf(stderr, "radiusd: The options -i and -p cannot be used individually.\n");
 		exit(1);
@@ -285,6 +281,30 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
+#ifdef HAVE_OPENSSL_CRYPTO_H
+	/*
+	 *	Initialize the OpenSSL library before calling any of its
+	 *	functions.
+	 */
+	SSL_library_init();
+	SSL_load_error_strings();
+
+	/*
+	 *	Mismatch between build time OpenSSL and linked SSL,
+	 *	better to die here than segfault later.
+	 */
+#ifdef ENABLE_OPENSSL_VERSION_CHECK
+	if (ssl_check_version(mainconfig.allow_vulnerable_openssl) < 0) {
+		exit(1);
+	}
+#endif
+#endif
+
+	/*  Load the modules AFTER doing SSL checks */
+	if (setup_modules(FALSE, mainconfig.config) < 0) {
+		return -1;
+	}
+
 	/* Set the panic action (if required) */
 	if (mainconfig.panic_action &&
 #ifndef NDEBUG
@@ -372,24 +392,68 @@ int main(int argc, char *argv[])
 	 */
 	radius_pid = getpid();
 
+	/*
+	 *	STDOUT & STDERR go to /dev/null, unless we have "-x",
+	 *	then STDOUT & STDERR go to the "-l log" destination.
+	 *
+	 *	The complexity here is because "-l log" can go to
+	 *	STDOUT or STDERR, too.
+	 */
 	if (mainconfig.radlog_dest == RADLOG_STDOUT) {
 		setlinebuf(stdout);
 		mainconfig.radlog_fd = STDOUT_FILENO;
-	} else if (debug_flag) {
-		dup2(devnull, STDOUT_FILENO);
-	}
 
-	if (mainconfig.radlog_dest == RADLOG_STDERR) {
-		setlinebuf(stdout);
+		/*
+		 *	If we're debugging, allow STDERR to go to
+		 *	STDOUT too, for executed programs,
+		 */
+		if (debug_flag) {
+			dup2(STDOUT_FILENO, STDERR_FILENO);
+		} else {
+			dup2(devnull, STDERR_FILENO);
+		}
+
+	} else if (mainconfig.radlog_dest == RADLOG_STDERR) {
+		setlinebuf(stderr);
 		mainconfig.radlog_fd = STDERR_FILENO;
-	} else {
-		dup2(devnull, STDERR_FILENO);
-	}
 
-	/* Libraries may write messages to stderr or stdout */
-	if (debug_flag) {
+		/*
+		 *	If we're debugging, allow STDOUT to go to
+		 *	STDERR too, for executed programs,
+		 */
+		if (debug_flag) {
+			dup2(STDERR_FILENO, STDOUT_FILENO);
+		} else {
+			dup2(devnull, STDOUT_FILENO);
+		}
+
+	} else if (mainconfig.radlog_dest == RADLOG_SYSLOG) {
+		/*
+		 *	Discard STDOUT and STDERR no matter what the
+		 *	status of debugging.  Syslog isn't a file
+		 *	descriptor, so we can't use it.
+		 */
+		dup2(devnull, STDOUT_FILENO);
+		dup2(devnull, STDERR_FILENO);
+
+	} else if (debug_flag) {
+		/*
+		 *	If we're debugging, allow STDOUT and STDERR to
+		 *	go to the log file.
+		 */
 		dup2(mainconfig.radlog_fd, STDOUT_FILENO);
 		dup2(mainconfig.radlog_fd, STDERR_FILENO);
+
+	} else {
+		/*
+		 *	Not debugging, and the log isn't STDOUT or
+		 *	STDERR.  Ensure that we move both of them to
+		 *	/dev/null, so that the calling terminal can
+		 *	exit, and the output from executed programs
+		 *	doesn't pollute STDOUT / STDERR.
+		 */
+		dup2(devnull, STDOUT_FILENO);
+		dup2(devnull, STDERR_FILENO);
 	}
 
 	close(devnull);
